///////////////////////////////////////////////////////////////////////////////
// FileDatagroupRecord.cpp
// Paul Wilson
// Oxford Metrics, May 2000
///////////////////////////////////////////////////////////////////////////////

#include "FileDatagroupRecord.h"
#include <limits>
#include <cassert>

VFileDatagroupRecord::VFileDatagroupRecord(SSHORT a_ID, EDataType a_Type, BYTE a_DataWidth, float a_FrameRate)
{
	assert(a_ID >= 0);
	assert(a_Type != EUndefined);
	assert(a_DataWidth > 0);
	assert(a_FrameRate > 0.0);

	m_GroupID = a_ID;
	m_DataType = a_Type;
	m_DataWidth = a_DataWidth;
	m_FrameRate = a_FrameRate;
}

VFileDatagroupRecord::VFileDatagroupRecord(BYTE * a_RawData, short a_sRecLen) 
:	VFileRecord(a_RawData, a_sRecLen)
{
	m_GroupID = -1;
	m_strDesc = "";
	m_DataType = EUndefined;
	m_DataWidth = 0;
	m_FrameRate = 0;
}

VFileDatagroupRecord::~VFileDatagroupRecord()
{
	// any specific clean-up first.. then call the base destructor
	//VFileRecord::~VFileRecord();
}

void VFileDatagroupRecord::SetGroupID(SSHORT a_ID)
{
	assert(a_ID >= 0);

	m_GroupID = a_ID;
	SetEdited();
}

void VFileDatagroupRecord::SetDescription(std::string a_strDesc)
{
	m_strDesc = a_strDesc;
	SetEdited();
}

void VFileDatagroupRecord::SetDataType(EDataType a_DataType)
{
	m_DataType = a_DataType;
	SetEdited();
}

void VFileDatagroupRecord::SetDataWidth(BYTE a_Width)
{
	m_DataWidth = a_Width;
	SetEdited();
}

void VFileDatagroupRecord::SetFrameRate(float a_FrameRate)
{
	assert(a_FrameRate > 0);

	m_FrameRate = a_FrameRate;
	SetEdited();
}

void VFileDatagroupRecord::DeleteLabel(int a_Index)
{
	assert(a_Index < m_Labels.size());

	std::vector<std::string>::iterator LabelIt = m_Labels.begin();

	for (int i = 0; i < a_Index; i++)
		LabelIt++;

	m_Labels.erase(LabelIt);
	SetEdited();
}

void VFileDatagroupRecord::AddLabel(std::string a_strNewLabel)
{
	m_Labels.push_back(a_strNewLabel);
	SetEdited();
}

void VFileDatagroupRecord::InsertLabel(std::string a_strNewLabel, int a_Index)
// Will assert if a_Index is the end of the list or beyond. To add to the end
// of the list use AddLabel
{
	assert(a_Index >= 0 && a_Index < m_Labels.size());

	std::vector<std::string>::iterator LabelIt = m_Labels.begin();

	for (int i = 0; i < a_Index; i++)
		LabelIt++;

	m_Labels.insert(LabelIt, a_strNewLabel);
	SetEdited();
}

void VFileDatagroupRecord::SetLabelAt(std::string a_strNewLabel, int a_Index)
{
	assert(a_Index >= 0 && a_Index < m_Labels.size());

	m_Labels[a_Index] = a_strNewLabel;
	SetEdited();
}

void VFileDatagroupRecord::DeleteDynamicRecord(int a_Index)
{
	std::vector<VFileDynamicRecord *>::iterator DynIt = m_DynRecs.begin();

	for (int i = 0; i < a_Index; i++)
		DynIt++;

	m_DynRecs.erase(DynIt);
}

void VFileDatagroupRecord::AddDynamicRecord(VFileDynamicRecord * a_pRec)
{
	assert(a_pRec != NULL);

	m_DynRecs.push_back(a_pRec);
}

void VFileDatagroupRecord::InsertDynamicRecord(VFileDynamicRecord * a_pRec, int a_Index)
{
	assert(a_pRec != NULL);
	assert(a_Index >= 0 && a_Index < m_DynRecs.size());

	std::vector<VFileDynamicRecord *>::iterator DynIt = m_DynRecs.begin();

	for (int i = 0; i < a_Index; i++)
		DynIt++;

	m_DynRecs.insert(DynIt, a_pRec);
}

void VFileDatagroupRecord::SetDynamicRecordAt(VFileDynamicRecord * a_pRec, int a_Index)
{
	assert(a_pRec != NULL);
	assert(a_Index >= 0 && a_Index < m_DynRecs.size());

	m_DynRecs[a_Index] = a_pRec;
}

int VFileDatagroupRecord::CalculateLength() const
{
	int DescLen = m_strDesc.length();
	
	if (DescLen > 0)
		DescLen++;	// Add one for null-terminator

	int Size = sizeof(SSHORT) +		// Group ID field
			   sizeof(BYTE) +		// Desc len field
			   DescLen +			// Description field
			   sizeof(BYTE) +		// Type field
			   sizeof(BYTE) +		// Width field
			   sizeof(float) +		// Frame rate
			   sizeof(SSHORT);		// Num DOFs

	// now add on the number of bytes needed to store the labels

	std::vector<std::string>::const_iterator LabelIt;
	
//	for (LabelIt = const_cast<std::vector<std::string>::iterator>(m_Labels.begin()); LabelIt != m_Labels.end(); LabelIt++)
	for (LabelIt = m_Labels.begin(); LabelIt != m_Labels.end(); LabelIt++)
	{
		Size += sizeof(BYTE);		// Label length field
		
		Size += LabelIt->length() + 1;	// +1 for null terminator
	}

	return Size;
}

void VFileDatagroupRecord::BuildRecordFromBytes()
{
	int ByteIndex = 0;

//	m_GroupID = (SSHORT)&(m_RawData[ByteIndex]);
	memcpy(&m_GroupID, &m_RawData[ByteIndex], sizeof(SSHORT));
	ByteIndex += sizeof(SSHORT);

	BYTE DescLen = m_RawData[ByteIndex];
	ByteIndex += sizeof(BYTE);

	if (DescLen > 0)
	{
		m_strDesc = (char *)&(m_RawData[ByteIndex]);	// Is this dodgy?
		ByteIndex += DescLen;
	}

	m_DataType = (EDataType)m_RawData[ByteIndex];
	ByteIndex += sizeof(BYTE);

	m_DataWidth = m_RawData[ByteIndex];
	ByteIndex += sizeof(BYTE);

//	m_FrameRate = (float)m_RawData[ByteIndex];
	memcpy(&m_FrameRate, &m_RawData[ByteIndex], sizeof(float));
	ByteIndex += sizeof(float);

	SSHORT NumDOFs;// = (SSHORT)&(m_RawData[ByteIndex]);
	memcpy(&NumDOFs, &m_RawData[ByteIndex], sizeof(SSHORT));
	ByteIndex += sizeof(SSHORT);

	m_Labels.clear();

	for (SSHORT i = 0; i < NumDOFs; i++)
	{
		BYTE LabelLen = m_RawData[ByteIndex];
		ByteIndex += sizeof(BYTE); 

		BYTE * szLabel = (BYTE *)&m_RawData[ByteIndex];
		ByteIndex += LabelLen;

		m_Labels.push_back((char *)szLabel);
	}
	
	assert(ByteIndex == GetLength());
}

SSHORT VFileDatagroupRecord::BuildBytesFromRecord()
{
	SSHORT OldSize = GetLength();
	SSHORT NewSize = CalculateLength();

	if (OldSize != NewSize)
	{
		delete [] m_RawData;

		m_RawData = new BYTE[NewSize];
	}

	assert(m_RawData);

	if (m_RawData)
	{
		// Okay, we've got the storage necessary, proceed..

		int ByteIndex = 0;

		memcpy(&m_RawData[ByteIndex], (BYTE *)&m_GroupID, sizeof(SSHORT));
		ByteIndex += sizeof(SSHORT);	

		assert(m_strDesc.length() + 1 < std::numeric_limits<BYTE>::max());	// Check this assert...

		int DescLen = m_strDesc.length();

		// Allow for null terminator if necessary

		if (DescLen)
			DescLen++;

		m_RawData[ByteIndex] = (BYTE)DescLen;
		ByteIndex += sizeof(BYTE);

		memcpy(&m_RawData[ByteIndex], (BYTE *)m_strDesc.c_str(), DescLen);
		ByteIndex += DescLen;

		assert(m_DataType < std::numeric_limits<BYTE>::max());	// Check this assert...
		m_RawData[ByteIndex] = (BYTE)m_DataType;
		ByteIndex += sizeof(BYTE);

		m_RawData[ByteIndex] = m_DataWidth;
		ByteIndex += sizeof(BYTE);

		memcpy(&m_RawData[ByteIndex], (BYTE *)&m_FrameRate, sizeof(float));
		ByteIndex += sizeof(float);

		SSHORT NumLabels = m_Labels.size();

		memcpy(&m_RawData[ByteIndex], (BYTE *)&NumLabels, sizeof(SSHORT));
		ByteIndex += sizeof(SSHORT);

		std::vector<std::string>::const_iterator LabelIt;

//		for (LabelIt = const_cast<std::vector<std::string>::iterator>(m_Labels.begin()); LabelIt != m_Labels.end(); LabelIt++)
		for (LabelIt = m_Labels.begin(); LabelIt != m_Labels.end(); LabelIt++)
		{		
			BYTE LabelLen = LabelIt->length() + 1;
			assert(LabelLen < std::numeric_limits<BYTE>::max());	// Check this assert...
			m_RawData[ByteIndex] = LabelLen;	// another one to check
			ByteIndex += sizeof(BYTE);

			memcpy(&m_RawData[ByteIndex], (BYTE *)LabelIt->c_str(), LabelLen);
			ByteIndex += LabelLen;
		}

		assert(ByteIndex == NewSize);

		SetLength(NewSize);
	}
	else
	{
		// Run out of memory...
		SetLength(0);
		NewSize = 0;
	}

	return NewSize;
}

std::string VFileDatagroupRecord::GetDataTypeAsString() const
{
	std::string strRet = "";

	switch (m_DataType)
	{
	case EBinary:
		strRet = "Binary";
		break;
	case EText:
		strRet = "Text";
		break;
	case EShort:
		strRet = "Short";
		break;
	case ELong:
		strRet = "Long";
		break;
	case EFloat:
		strRet = "Float";
		break;
	case EDouble:
		strRet = "Double";
		break;
	case EBoolean:
		strRet = "Boolean";
		break;
	default:
		assert(false);
		break;
	}

	return strRet;
}

