#include "Body.h"
#include <math.h>
#include <maya/MQuaternion.h>
#include <maya/MEulerRotation.h>
#include <maya/MMatrix.h>

#define PI 3.14159265358979323846

VMayaBody::VMayaBody(std::string a_strBodyName, int a_GroupID, EOrient a_Orient) 
{ 
	m_strBodyName = a_strBodyName; 
	m_WorldUp = a_Orient;
	m_DatagroupID = a_GroupID;

	for (int i = 0; i < 6; i++)
	{
		m_DofIndices[i] = -1;
		m_pAnimCurves[i] = NULL;
	}

	m_MayaFrameRate = GetMayaFrameRate();
	m_DistUnit = MDistance::internalUnit();
	m_bFirstEuler = true;
}

VMayaBody::~VMayaBody()
{
	for (int i = 0; i < 6; i++)
	{
		if (m_pAnimCurves[i] != NULL)
		{
			delete m_pAnimCurves[i];
			m_pAnimCurves[i] = NULL;
		}
	}
}

VMayaBody::EDof VMayaBody::ConvertDofSuffix(std::string a_strPre, std::string & a_rstrPost)
{
	EDof eRet = EUndefDof;

	a_strPre = TrimLeft(a_strPre);
	a_strPre = TrimRight(a_strPre);

	if (_stricmp(a_strPre.c_str(), V_POSX) == 0)
	{
		a_rstrPost = MAYA_POSX;
		eRet = ETx;
	}
	else if (_stricmp(a_strPre.c_str(), V_POSY) == 0)
	{
		a_rstrPost = MAYA_POSY;
		eRet = ETy;
	}
	else if (_stricmp(a_strPre.c_str(), V_POSZ) == 0)
	{
		a_rstrPost = MAYA_POSZ;
		eRet = ETz;
	}
	else if (_stricmp(a_strPre.c_str(), V_ROTX) == 0)
	{
		a_rstrPost = MAYA_ROTX;
		eRet = ERx;
	}
	else if (_stricmp(a_strPre.c_str(), V_ROTY) == 0)
	{
		a_rstrPost = MAYA_ROTY;
		eRet = ERy;
	}
	else if (_stricmp(a_strPre.c_str(), V_ROTZ) == 0)
	{
		a_rstrPost = MAYA_ROTZ;
		eRet = ERz;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALROTX) == 0)
	{
		a_rstrPost = MAYA_ROTX;
		eRet = ERx;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALROTY) == 0)
	{
		a_rstrPost = MAYA_ROTY;
		eRet = ERy;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALROTZ) == 0)
	{
		a_rstrPost = MAYA_ROTZ;
		eRet = ERz;
	}
	else if (_stricmp(a_strPre.c_str(), V_TRANSX) == 0)
	{
		a_rstrPost = MAYA_POSX;
		eRet = ETx;
	}
	else if (_stricmp(a_strPre.c_str(), V_TRANSY) == 0)
	{
		a_rstrPost = MAYA_POSY;
		eRet = ETy;
	}
	else if (_stricmp(a_strPre.c_str(), V_TRANSZ) == 0)
	{
		a_rstrPost = MAYA_POSZ;
		eRet = ETz;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALTRANSX) == 0)
	{
		a_rstrPost = MAYA_POSX;
		eRet = ETx;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALTRANSY) == 0)
	{
		a_rstrPost = MAYA_POSY;
		eRet = ETy;
	}
	else if (_stricmp(a_strPre.c_str(), V_LOCALTRANSZ) == 0)
	{
		a_rstrPost = MAYA_POSZ;
		eRet = ETz;
	}
	else
	{
		a_rstrPost = "";
	}

	return eRet;
}

void VMayaBody::SetDofIndex(EDof a_eComponent, int a_Index)
{
	assert(a_eComponent != EUndefDof);
	assert(a_Index >= 0);

	if (a_eComponent != EUndefDof)
	{
		// Check this hasn't already been set
		assert(m_DofIndices[a_eComponent] == -1);	

		m_DofIndices[a_eComponent] = a_Index; 
	}
}

void VMayaBody::SetAnimCurve(EDof a_eComponent, MFnAnimCurve * a_pCurve)
{
	assert(a_eComponent != EUndefDof);
	assert(a_pCurve != NULL);

	if (a_eComponent != EUndefDof)
	{
		assert(m_pAnimCurves[a_eComponent] == NULL);

		m_pAnimCurves[a_eComponent] = a_pCurve;
	}
}
/*
  kGames 
15 frames per second  
  kFilm 
24 frames per second  
  kPALFrame 
25 frames per second  
  kNTSCFrame 
30 frames per second  
  kShowScan 
twice the speed of film (48 frames per second)  
  kPALField 
twice the speed of PAL (50 frames per second)  
  kNTSCField 
twice the speed of NTSC (60 frames per second) 
*/

void VMayaBody::SetKeyframe(VFileDynamicRecord & a_rRecord, float a_FrameRate)
{
	// If this is not the correct datagroup then it aint nothing to do with us

	if (a_rRecord.GetDatagroupID() != m_DatagroupID)
		return;

//	MTime::Unit timeUnit = MTime::uiUnit();

	float FrameNum = a_rRecord.GetFrameNumber();
	
	// Scale all the motion to maya's internal frame rate

	MTime timeData(FrameNum * m_MayaFrameRate / a_FrameRate, MTime::uiUnit());

	// First do translation...

	double DofValue;

	if (m_DofIndices[ETx] != -1)
	{
		if (a_rRecord.GetDofValAsDouble(m_DofIndices[ETx], DofValue))
		{
			assert(m_pAnimCurves[ETx] != NULL);

			if (m_pAnimCurves[ETx])
			{
				// Scale the value...

				ConvertToInternalUnits(DofValue);
				MStatus TXStatus = m_pAnimCurves[ETx]->addKeyframe(timeData, DofValue, MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);

				assert(TXStatus == MStatus::kSuccess);
			}
		}
	}

	if (m_DofIndices[ETy] != -1)
	{
		if (a_rRecord.GetDofValAsDouble(m_DofIndices[ETy], DofValue))
		{
			assert(m_pAnimCurves[ETy] != NULL);

			if (m_pAnimCurves[ETy])
			{
				// Scale the value...

				ConvertToInternalUnits(DofValue);
				MStatus TYStatus = m_pAnimCurves[ETy]->addKeyframe(timeData, DofValue, MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);

				assert(TYStatus == MStatus::kSuccess);
			}
		}
	}

	if (m_DofIndices[ETz] != -1)
	{
		if (a_rRecord.GetDofValAsDouble(m_DofIndices[ETz], DofValue))
		{
			assert(m_pAnimCurves[ETz] != NULL);

			if (m_pAnimCurves[ETz])
			{
				// Scale the value...

				ConvertToInternalUnits(DofValue);
				MStatus TZStatus = m_pAnimCurves[ETz]->addKeyframe(timeData, DofValue, MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);

				assert(TZStatus == MStatus::kSuccess);
			}
		}
	}

	double AngleAxis[3] = {0.0};
	double Quaternion[4] = {0.0};
	
	// Need ALL rotational components to make a sensible rotation since this is angle-axis

	if (m_DofIndices[ERx] != -1 && m_DofIndices[ERy] != -1 && m_DofIndices[ERz] != -1)
	{
		if (!a_rRecord.GetDofValAsDouble(m_DofIndices[ERx], AngleAxis[ERx]))
			return;

		if (!a_rRecord.GetDofValAsDouble(m_DofIndices[ERy], AngleAxis[ERy]))
			return;

		if (!a_rRecord.GetDofValAsDouble(m_DofIndices[ERz], AngleAxis[ERz]))
			return;

		assert(m_pAnimCurves[ERx] != NULL);
		assert(m_pAnimCurves[ERy] != NULL);
		assert(m_pAnimCurves[ERz] != NULL);

		if (m_pAnimCurves[ERx] && m_pAnimCurves[ERy] && m_pAnimCurves[ERz])
		{
			// There are no Angle-axis classes within Maya... have to do the 
			// conversion ourselves

			double CurFrame = timeData.value();

			ConvertAngleAxisToQuat(Quaternion, AngleAxis);
	
			// Now use Maya's classes to convert to Euler

			MQuaternion MayaQuat(Quaternion);
			
			MEulerRotation KeyFrameEuler;
			KeyFrameEuler = MayaQuat;

			if (m_bFirstEuler)
				m_bFirstEuler = false;
			else
				KeyFrameEuler.setToClosestSolution(m_LastEuler);
				
			m_LastEuler = KeyFrameEuler;

			MStatus RXStatus = m_pAnimCurves[ERx]->addKeyframe(timeData, KeyFrameEuler[ERx], MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);
			MStatus RYStatus = m_pAnimCurves[ERy]->addKeyframe(timeData, KeyFrameEuler[ERy], MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);
			MStatus RZStatus = m_pAnimCurves[ERz]->addKeyframe(timeData, KeyFrameEuler[ERz], MFnAnimCurve::kTangentLinear, MFnAnimCurve::kTangentLinear);

			assert(RXStatus == MStatus::kSuccess);
			assert(RYStatus == MStatus::kSuccess);
			assert(RZStatus == MStatus::kSuccess);
		}
	}
}

void VMayaBody::ConvertAngleAxisToQuat(double a_q[4], const double a_r[3])
{
	double len, tmp;

	len = sqrt(a_r[0] * a_r[0] + a_r[1] * a_r[1] + a_r[2] * a_r[2]);
	a_q[3] = cos(len / 2.0);
	tmp = sin(len / 2.0);

	if (len < ConvSmallNum) 
	{
		a_q[0] = a_r[0];
		a_q[1] = a_r[1];
		a_q[2] = a_r[2];
	} 
	else 
	{
		a_q[0] = a_r[0] * tmp/len;
		a_q[1] = a_r[1] * tmp/len;
		a_q[2] = a_r[2] * tmp/len;
	}
}

void VMayaBody::ConvertAngleAxisToMatrix(double a_M[4][4], const double a_r[3])
{
	double angle, c, s, x, y, z;

	angle = sqrt(a_r[0]*a_r[0] + a_r[1]*a_r[1] + a_r[2]*a_r[2]);

	if (angle < ConvSmallNum) 
	{
		a_M[0][0] = a_M[1][1] = a_M[2][2] = 1.0;
		a_M[1][0] = a_M[2][0] = a_M[0][1] = a_M[2][1] = a_M[0][2] = a_M[1][2] = 0.0;
		return;
	} 
	else 
	{
		x = a_r[0]/angle;
		y = a_r[1]/angle;
		z = a_r[2]/angle;
	}

	c = cos(angle);
	s = sin(angle);

	a_M[0][0] = c + (1-c)*x*x;
	a_M[1][0] =     (1-c)*x*y + s*(-z);
	a_M[2][0] =     (1-c)*x*z + s*y;
	a_M[0][1] =     (1-c)*y*x + s*z;
	a_M[1][1] = c + (1-c)*y*y;
	a_M[2][1] =     (1-c)*y*z + s*(-x);
	a_M[0][2] =     (1-c)*z*x + s*(-y);
	a_M[1][2] =     (1-c)*z*y + s*x;
	a_M[2][2] = c + (1-c)*z*z;
}

/*
  kGames 
15 frames per second  
  kFilm 
24 frames per second  
  kPALFrame 
25 frames per second  
  kNTSCFrame 
30 frames per second  
  kShowScan 
twice the speed of film (48 frames per second)  
  kPALField 
twice the speed of PAL (50 frames per second)  
  kNTSCField 
twice the speed of NTSC (60 frames per second) 
*/

float GetMayaFrameRate()
{
	MTime::Unit Units = MTime::uiUnit();
	float FrameRate = 1.0f;

	switch (Units)
	{
	case MTime::kGames:
		FrameRate = 15.0f;
		break;
	case MTime::kFilm:
		FrameRate = 24.0f;
		break;
	case MTime::kPALFrame:
		FrameRate = 25.0f;
		break;
	case MTime::kNTSCFrame:
		FrameRate = 30.0f;
		break;
	case MTime::kShowScan:
		FrameRate = 48.0f;
		break;
	case MTime::kPALField:
		FrameRate = 50.0f;
		break;
	case MTime::kNTSCField:
		FrameRate = 60.0f;
		break;
	}

	return FrameRate;
}

void VMayaBody::ConvertToInternalUnits(double & a_Value)
{
	MDistance RealVal;
	RealVal.setUnit(MDistance::kMillimeters);
	RealVal.setValue(a_Value);

	a_Value = RealVal.as(m_DistUnit);
}

void VMayaBody::ConvertToInternalUnits(float & a_Value)
{
	MDistance RealVal;
	RealVal.setUnit(MDistance::kMillimeters);
	RealVal.setValue(a_Value);

	a_Value = RealVal.as(m_DistUnit);
}

std::string TrimLeft(std::string a_strLabel)
{
	int WhiteCount = 0;
	int i = 0;

	while (a_strLabel[i] == ' ' || a_strLabel[i] == '\t' || a_strLabel[i] == '\r' || a_strLabel[i] == '\n')
	{
		WhiteCount++;
		i++;
	}

	return a_strLabel.substr(WhiteCount);
}

std::string TrimRight(std::string a_strLabel)
{
	int WhiteCount = 0;
	int Len = a_strLabel.length();
	int i = Len - 1;

	while (a_strLabel[i] == ' ' || a_strLabel[i] == '\t' || a_strLabel[i] == '\r' || a_strLabel[i] == '\n')
	{
		WhiteCount++;
		i--;
	}

	return a_strLabel.substr(0, Len - WhiteCount);
}

