///////////////////////////////////////////////////////////////////////////////
// DatagroupSectionHandler.cpp
// Paul Wilson
// Oxford Metrics, May 2000
///////////////////////////////////////////////////////////////////////////////

//#pragma once

#include "DatagroupSectionHandler.h"
#include "FileDatagroupSection.h"

#include <cassert>

VDatagroupSectionHandler::VDatagroupSectionHandler(VExtendedFileWrapper & a_rFileWrapper)
: VSectionHandler(a_rFileWrapper)
{

}

VDatagroupSectionHandler::~VDatagroupSectionHandler()
{

}

bool VDatagroupSectionHandler::CanHandle(std::string a_strSectionName)
{
	return _stricmp(a_strSectionName.c_str(), DATAGROUP_SECTION) == 0;
}

VSectionHandler::EParseResult VDatagroupSectionHandler::ReadSection(VFileContents & a_rContents, SLONG a_SectionLength)
{
	VSectionHandler::EParseResult eResult = ENoError;
	VFileDatagroupSection * pNewSection = new VFileDatagroupSection(a_SectionLength);

	if (pNewSection != NULL)
	{
		SSHORT RecLen = 0;

		if (m_rFile.Read(&RecLen, sizeof(SSHORT)))
		{
			// loop until we find the Section terminator

			while (RecLen != 0)
			{
				eResult = ReadRecord(pNewSection, RecLen);

				if (eResult == EReadError)
				{
					// Okay.. that record was the last one (and it was incomplete).
					// Have to make do...

					break;
				}
				else if (eResult == EOutOfMemory)
				{
					// This is a fairly serious problem... better cleanup

					delete pNewSection;
					return EOutOfMemory;
				}

				if (!m_rFile.Read(&RecLen, sizeof(SSHORT)))
					return EReadError;
			}

			// Right.. we've read all the records... now figure out how big the 
			// trailing buffer is and skip it...

			int NumToSkip = a_SectionLength - pNewSection->GetDataLength();
			m_rFile.Seek(NumToSkip, SEEK_CUR);

		}
		else
			return EReadError;
	}
	else
		return EOutOfMemory;

	a_rContents.AddSection(pNewSection);

	return ENoError;
}

VSectionHandler::EParseResult VDatagroupSectionHandler::ReadRecord(VFileDatagroupSection * a_pSection, SSHORT a_RecLen)
{
	VFileDatagroupRecord * pNewRecord = NULL;

	VSectionHandler::EParseResult eRetVal = ENoError;
	BYTE * pNewBuffer = new BYTE[a_RecLen];

	if (pNewBuffer != NULL)
	{
		if (m_rFile.Read(pNewBuffer, a_RecLen))
		{
			pNewRecord = new VFileDatagroupRecord(pNewBuffer, a_RecLen);

			if (pNewRecord == NULL)
			{
				eRetVal = EOutOfMemory;
				delete [] pNewBuffer;
			}
			else
			{
				// Everything is fine...

				pNewRecord->BuildRecordFromBytes();
				a_pSection->AddRecord(pNewRecord);
			}
		}
		else
		{
			eRetVal = EReadError;

			// Half a record aint no use to us...	

			delete [] pNewBuffer;
		}
	}
	else
	{
		eRetVal = EOutOfMemory;
	}

	return eRetVal;
}

void VDatagroupSectionHandler::WriteSection(VFileSection * a_pSection)
{
	assert(a_pSection != NULL);
	VFileDatagroupSection * pDGSection = dynamic_cast<VFileDatagroupSection *>(a_pSection);
	assert(pDGSection != NULL);

	a_pSection->PrepareForWrite();
	SLONG TotalLen = a_pSection->GetDataLength() + BUFFER_SIZE;

	WriteSectionHeader(TotalLen, a_pSection->GetName());
		
	// Write all records

	int NumRecs = pDGSection->GetNumRecords();

	for (int i = 0; i < NumRecs; i++)
	{
		const VFileDatagroupRecord * pRecord = pDGSection->GetRecordAt(i);
		WriteRecord(const_cast<VFileDatagroupRecord *>(pRecord));
	}

	WriteSectionTerminator();

	// Finally set the new section length

	a_pSection->SetLength(TotalLen);
}

bool VDatagroupSectionHandler::WriteSectionHeader(SLONG a_SectionLength, std::string a_strName)
{
	bool bRetVal = false;

	bRetVal = m_rFile.Write(&a_SectionLength, sizeof(SLONG));

	BYTE szName[SECTION_NAME_SIZE] = {0};
	memcpy(szName, a_strName.c_str(), SECTION_NAME_SIZE);
	szName[SECTION_NAME_SIZE - 1] = 0;	// Ensure null-termination

	if (bRetVal == true)
		bRetVal = m_rFile.Write(szName, SECTION_NAME_SIZE);

	return bRetVal;
}

void VDatagroupSectionHandler::WriteSectionTerminator()
{
	SSHORT NullRecord = 0;

	m_rFile.Write(&NullRecord, sizeof(SSHORT));
}

/*
SLONG VDatagroupSectionHandler::WriteSectionData(VFileDatagroupSection * a_pSection)
{
	// Write all records

	int NumRecs = a_pSection->GetNumRecords();

	// There's one header per record... plus one for the section terminator, so...

	SLONG TotalLen = sizeof(SSHORT) * (NumRecs + 1);

	for (int i = 0; i < NumRecs; i++)
	{
		VFileDatagroupRecord * pRecord = a_pSection->GetRecordAt(i);
		pRecord->BuildBytesFromRecord();

		WriteRecord(pRecord);
		TotalLen += pRecord->GetLength();
	}

	// Write the section terminator

	SSHORT SectionTerminator = 0;
	m_rFile.Write(&SectionTerminator, sizeof(SSHORT));
	TotalLen += sizeof(SSHORT);

	return TotalLen;
}
*/
SLONG VDatagroupSectionHandler::WriteSectionBuffer()
{
	return 0;
		
	// Write a buffer
//	BYTE TrailingBuffer[BUFFER_SIZE] = {0};

//	return m_rFile.Write(TrailingBuffer, sizeof(TrailingBuffer));
}

void VDatagroupSectionHandler::WriteRecord(VFileDatagroupRecord * a_pRecord)
{
	SSHORT RecLen = a_pRecord->GetLength();
	const BYTE * pBuffer = a_pRecord->GetDataBuffer();

	// First write reclen field...
	m_rFile.Write(&RecLen, sizeof(SSHORT));
	
	// then the data
	m_rFile.Write(pBuffer, RecLen);
}
/*
bool VDatagroupSectionHandler::RewriteSection(VFileSection * a_pSection)
{
	bool bRetVal = false;

	assert(a_pSection != NULL);
	VFileDatagroupSection * pDGSection = dynamic_cast<VFileDatagroupSection *>(a_pSection);
	assert(pDGSection != NULL);

	SLONG TotalSize = a_pSection->GetTotalLength();

	if (TotalSize <= a_pSection->GetTotalLength())
	{
		// Yes it will... this is the easy case - just rewrite the section 
		// data and leave the buffer intact.

		WriteSectionData(pDGSection);
	}
	else
	{
		// Need to rewrite the rest of the file
	}

	return bRetVal;
}
*/

bool VDatagroupSectionHandler::RewriteSection(VFileSection * a_pSection)
{
	assert(false);

	return false;
}

